#!/usr/bin/python3

import numpy as np
import matplotlib.pyplot as plt
import sys
from scipy import ndimage
from skimage import feature
from skimage import filters
from skimage.feature import corner_harris, corner_subpix, corner_peaks

def detect_corner(img, windowx = 2, windowy = 2):
    (n,m) = img.shape
    corners = np.zeros((n,m), dtype=np.bool)
    nr_cuts = 0
    for i in range(0, n):
        print("Row: {}".format(i))
        for j in range(0, m):
            corner = False
            for k in range(-windowy, windowy):
                nr_cuts = 0
                for l in range(-windowx, windowx):
                    y  = (i+k)%n
                    x1  = (j+l)%m
                    x2 = (j+l+1)%m
                    if img[y,x1] ^ img[y,x2]:
                        nr_cuts += 1
                if nr_cuts > 1:
                    corner = True
#                    print("{}x{} is corner".format(j,i))
                    break
            corners[i,j] = corner
    return corners

def produce_poi_list(img):
    r = []
    (n,m) = img.shape
    for i in range(0, n):
        row = []
        for j in range(0, m):
            if img[i,j]:
                row.append((float(j)/m,float(i)/n))
        r.append(row)

    return r

# not time-effective comparision
# TODO compare only rows in the r window
def compare_poi_list(pl1, pl2, r):
    matches = []
    for pr1 in pl1:
        for (x1,y1) in pr1:
            for pr2 in pl2:
                for (x2,y2) in pr2:
                    if r * r > ((x1 - x2)*(x1 - x2) + (y1-y2)*(y1-y2)):
                        matches.append(((x1,y1),(x2,y2)))
    return matches

file_path = sys.argv[1]
print("File {} to be analysed".format(file_path))

img = ndimage.imread(file_path)
img_gray = 0.21 * img[:,:,0] + 0.72 * img[:,:,1] + 0.07 * img[:,:,2]

n,m = img_gray.shape

print("Width {}, Height {}".format(m,n))

otsu_lvl = filters.threshold_otsu(img_gray)
seg_shoe = img_gray <= otsu_lvl
img_shoe_selected = seg_shoe * img_gray
data_selected = img_gray[seg_shoe]
shoe_otsu_lvl = filters.threshold_otsu(data_selected)
img_sole_segs1 = img_shoe_selected <= shoe_otsu_lvl
img_sole_segs = ndimage.median_filter(img_sole_segs1, 5)
img_sole_segs_nr = np.array(img_sole_segs, np.int)
k = np.ones((11, 11))
img_sole_seg_count = ndimage.convolve(img_sole_segs_nr, k)
m1 = img_sole_seg_count >= 80
m2 = True ^ img_sole_segs
# m2 = img_sole_seg_count <= 1
m3 = m1 * m2
# m3i = np.array(m3, np.int)

corners = detect_corner(img_sole_segs)
poi_list = produce_poi_list(corners)
print(poi_list)

m3i = np.array(corners, np.int)

# color_red = np.ones((n,m,3)) * np.array([255, 0, 0])
color_red = 255
img_annotated = img_gray * (1 - m3i) + m3i * color_red
# img_annotated = img_gray


# plt.imshow(m1 * m2, cmap="Greys_r")
# plt.imshow(img_annotated, cmap="Greys_r")

# coords = corner_peaks(corner_harris(img_gray), min_distance=5)
# coords_subpix = corner_subpix(img_gray, coords, window_size=13)

fig = plt.figure()
ax1 = fig.add_subplot(2,1,1)
ax1.imshow(img_sole_segs, cmap="Greys_r")

ax2 = fig.add_subplot(2,1,2)
ax2.imshow(img_annotated, cmap="Greys_r")
# ax2.plot(coords[:, 1], coords[:, 0], '.b', markersize=3)
# ax2.plot(coords_subpix[:, 1], coords_subpix[:, 0], '+r', markersize=15)

plt.show()

plt.imsave("sole_seg.png", img_sole_segs)

